import logging
import logging.handlers
import os
#import warnings

#class Singleton(type):
#    _instances = {}
#    def __call__(cls, *args, **kwargs):
#        if cls not in cls._instances:
#            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
#        return cls._instances[cls]

#class Logger(metaclass=Singleton):
class Logger():
    def __init__(self):
        level = os.environ.get('LPSDS_LOG_LEVEL', 'DEBUG')
        svc_name = os.environ.get('LPSDS_LOG_SVC_NAME', 'job')
        logLevel = getattr(logging, level)
#        warnings.filterwarnings('default')
#        logging.captureWarnings(True)
        self.logger = logging.getLogger(svc_name)
        self.logger.setLevel(logLevel)

        stderr_log_handler = logging.StreamHandler()
        stderr_log_handler.setLevel(logLevel)
        stderr_log_handler.setFormatter(logging.Formatter('%(asctime)s - %(module)s - %(levelname)s: %(message)s'))
        self.logger.addHandler(stderr_log_handler)


log = Logger().logger



import pytest
import copy
import numpy as np
import pandas as pd
from lpsds.utils import ObjectView
from lpsds.operation import get_operation_model, get_fold_data, get_staged_metrics


class FoldDataBase:
    @pytest.fixture
    def cv_model(self):
        return dict(
            test_accuracy=np.array([3,9,6,8,4]), #max ids = 1
            test_mse=np.array([3,9,6,2,4]), #min idx = 3
            estimator=['model_40', 'model_41', 'model_42', 'model_43', 'model_44']
        )
    
    @pytest.fixture
    def cv_splits(self):
        return [
            ([2,3,4,5,6,7,8,9], [0,1]), #fold 0
            ([0,1,4,5,6,7,8,9], [2,3]), #fold 1
            ([0,1,2,3,6,7,8,9], [4,5]), #fold 2
            ([0,1,2,3,4,5,8,9], [6,7]), #fold 3
            ([0,1,2,3,4,5,6,7], [8,9]), #fold 4
        ]
    
    @pytest.fixture
    def X(self):
        ret = pd.DataFrame(columns=['x1', 'x2'])
        ret.loc[len(ret)] = 10,20
        ret.loc[len(ret)] = 11,21
        ret.loc[len(ret)] = 12,22
        ret.loc[len(ret)] = 13,23
        ret.loc[len(ret)] = 14,24
        ret.loc[len(ret)] = 15,25
        ret.loc[len(ret)] = 16,26
        ret.loc[len(ret)] = 17,27
        ret.loc[len(ret)] = 18,28
        ret.loc[len(ret)] = 19,29
        return ret
    
    @pytest.fixture
    def y_true(self):
        return pd.Series([
            30,
            31,
            32,
            33,
            34,
            35,
            36,
            37,
            38,
            39,
        ])


class TestGetFoldData(FoldDataBase):

    def test_x_trn(self, cv_model, cv_splits, X, y_true):
        x_trn, _, _, _, _ = get_fold_data(cv_model, cv_splits, X, y_true, 1)
        assert x_trn.shape[0] == 8
        assert x_trn.shape[1] == 2
        assert x_trn.x1.iloc[0] == 10
        assert x_trn.x2.iloc[0] == 20
        assert x_trn.x1.iloc[1] == 11
        assert x_trn.x2.iloc[1] == 21
        assert x_trn.x1.iloc[2] == 14
        assert x_trn.x2.iloc[2] == 24
        assert x_trn.x1.iloc[3] == 15
        assert x_trn.x2.iloc[3] == 25
        assert x_trn.x1.iloc[4] == 16
        assert x_trn.x2.iloc[4] == 26
        assert x_trn.x1.iloc[5] == 17
        assert x_trn.x2.iloc[5] == 27
        assert x_trn.x1.iloc[6] == 18
        assert x_trn.x2.iloc[6] == 28
        assert x_trn.x1.iloc[7] == 19
        assert x_trn.x2.iloc[7] == 29


    def test_x_tst(self, cv_model, cv_splits, X, y_true):
        _, x_tst, _, _, _ = get_fold_data(cv_model, cv_splits, X, y_true, 1)
        assert x_tst.shape[0] == 2
        assert x_tst.shape[1] == 2
        assert x_tst.x1.iloc[0] == 12
        assert x_tst.x2.iloc[0] == 22
        assert x_tst.x1.iloc[1] == 13
        assert x_tst.x2.iloc[1] == 23


    def test_y_trn(self, cv_model, cv_splits, X, y_true):
        _, _, y_trn, _, _ = get_fold_data(cv_model, cv_splits, X, y_true, 1)
        assert y_trn.shape[0] == 8
        assert y_trn.iloc[0] == 30
        assert y_trn.iloc[1] == 31
        assert y_trn.iloc[2] == 34
        assert y_trn.iloc[3] == 35
        assert y_trn.iloc[4] == 36
        assert y_trn.iloc[5] == 37
        assert y_trn.iloc[6] == 38
        assert y_trn.iloc[7] == 39


    def test_y_tst(self, cv_model, cv_splits, X, y_true):
        _, _, _, y_tst, _ = get_fold_data(cv_model, cv_splits, X, y_true, 1)
        assert y_tst.shape[0] == 2
        assert y_tst.iloc[0] == 32
        assert y_tst.iloc[1] == 33


    def test_model(self, cv_model, cv_splits, X, y_true):
        _, _, _, _, model = get_fold_data(cv_model, cv_splits, X, y_true, 1)
        assert model == 'model_41'


    def test_numpy_data_x_trn(self, cv_model, cv_splits, X, y_true):
        x_trn, _, _, _, _ = get_fold_data(cv_model, cv_splits, X.to_numpy(), y_true.to_numpy(), 2)
        assert x_trn.shape[0] == 8
        assert x_trn.shape[1] == 2
        assert x_trn[0,0] == 10
        assert x_trn[0,1] == 20
        assert x_trn[1,0] == 11
        assert x_trn[1,1] == 21
        assert x_trn[2,0] == 12
        assert x_trn[2,1] == 22
        assert x_trn[3,0] == 13
        assert x_trn[3,1] == 23
        assert x_trn[4,0] == 16
        assert x_trn[4,1] == 26
        assert x_trn[5,0] == 17
        assert x_trn[5,1] == 27
        assert x_trn[6,0] == 18
        assert x_trn[6,1] == 28
        assert x_trn[7,0] == 19
        assert x_trn[7,1] == 29


    def test_numpy_data_x_test(self, cv_model, cv_splits, X, y_true):
        _, x_tst, _, _, _ = get_fold_data(cv_model, cv_splits, X.to_numpy(), y_true.to_numpy(), 2)
        assert x_tst.shape[0] == 2
        assert x_tst.shape[1] == 2
        assert x_tst[0,0] == 14
        assert x_tst[0,1] == 24
        assert x_tst[1,0] == 15
        assert x_tst[1,1] == 25



    def test_numpy_data_y_trn(self, cv_model, cv_splits, X, y_true):
        _, _, y_trn, _, _ = get_fold_data(cv_model, cv_splits, X.to_numpy(), y_true.to_numpy(), 2)
        assert y_trn.shape[0] == 8
        assert y_trn[0] == 30
        assert y_trn[1] == 31
        assert y_trn[2] == 32
        assert y_trn[3] == 33
        assert y_trn[4] == 36
        assert y_trn[5] == 37
        assert y_trn[6] == 38
        assert y_trn[7] == 39


    def test_numpy_data_y_test(self, cv_model, cv_splits, X, y_true):
        _, _, _, y_tst, _ = get_fold_data(cv_model, cv_splits, X.to_numpy(), y_true.to_numpy(), 2)
        assert y_tst.shape[0] == 2
        assert y_tst[0] == 34
        assert y_tst[1] == 35



class TestGetOperationModel(FoldDataBase):

    @pytest.mark.parametrize(("metric", "less_is_better", 'target'), [ 
                                                    ('test_accuracy', False, 1),
                                                    ('test_mse', True, 3),
                                                ])
    def test_best_fold_idx(self, cv_model, cv_splits, X, y_true, metric, less_is_better, target):
        best_fold, _, _, _ = get_operation_model(cv_model, cv_splits, X, y_true, metric, less_is_better)
        assert best_fold == target


    @pytest.mark.parametrize(("metric", "less_is_better", 'target'), [ 
                                                    ('test_accuracy', False, [12,22,13,23]),
                                                    ('test_mse', True, [16,26,17,27]),
                                                ])
    def test_best_x(self, cv_model, cv_splits, X, y_true, metric, less_is_better, target):
        _, best_x, _, _ = get_operation_model(cv_model, cv_splits, X, y_true, metric, less_is_better)
        assert best_x.shape[0] == 2
        assert best_x.shape[1] == 2
        assert best_x.x1.iloc[0] == target[0]
        assert best_x.x2.iloc[0] == target[1]
        assert best_x.x1.iloc[1] == target[2]
        assert best_x.x2.iloc[1] == target[3]


    @pytest.mark.parametrize(("metric", "less_is_better", 'target'), [ 
                                                    ('test_accuracy', False, [32,33]),
                                                    ('test_mse', True, [36,37]),
                                                ])
    def test_best_y(self, cv_model, cv_splits, X, y_true, metric, less_is_better, target):
        _, _, best_y, _ = get_operation_model(cv_model, cv_splits, X, y_true, metric, less_is_better)
        assert best_y.shape[0] == 2
        assert best_y.iloc[0] == target[0]
        assert best_y.iloc[1] == target[1]


    @pytest.mark.parametrize(("metric", "less_is_better", 'target'), [ 
                                                    ('test_accuracy', False, 'model_41'),
                                                    ('test_mse', True, 'model_43'),
                                                ])
    def test_best_model(self, cv_model, cv_splits, X, y_true, metric, less_is_better, target):
        _, _, _, model = get_operation_model(cv_model, cv_splits, X, y_true, metric, less_is_better)
        assert model == target


    def test_numpy_data_x(self, cv_model, cv_splits, X, y_true):
        _, best_x, _, _ = get_operation_model(cv_model, cv_splits, X.to_numpy(), y_true, 'test_accuracy')
        assert best_x.shape[0] == 2
        assert best_x.shape[1] == 2
        assert best_x[0,0] == 12
        assert best_x[0,1] == 22
        assert best_x[1,0] == 13
        assert best_x[1,1] == 23


    @pytest.mark.parametrize(("y"), [
                                        (pd.Series([30,31,32,33,34,35,36,37,38,39]).to_numpy()),
                                        (np.array([30,31,32,33,34,35,36,37,38,39])),
                                        (np.array([30,31,32,33,34,35,36,37,38,39]).reshape(-1,1)),
                                    ])
    def test_numpy_data_y(self, cv_model, cv_splits, X, y):
        _, _, best_y, _ = get_operation_model(cv_model, cv_splits, X, y, 'test_accuracy')
        assert best_y.shape[0] == 2
        assert best_y[0] == 32
        assert best_y[1] == 33



class TestGetStagedMetrics:

    class FakePredictor:
        def staged_predict(self, X):
            X = X.copy()
            for g, r in enumerate(X.iterrows()):
                yield (g+1) * r[1].to_numpy()
    
    class FakePreProcessing:
        def __init__(self, offset):
            self.offset = offset

        def transform(self, X):
            return X - self.offset*np.ones(X.shape)

    def pipeline(self, offset):
        pipe = ObjectView()
        pipe.steps = [
            ('pre_proc', TestGetStagedMetrics.FakePreProcessing(offset)),
            ('estimator', TestGetStagedMetrics.FakePredictor())
        ]
        return pipe

    @pytest.fixture
    def cv_splits(self):
        return [
            ([3,4,5], [0,1,2]), #fold 0
            ([0,1,2], [3,4,5]), #fold 1
        ]
    
    @pytest.fixture
    def X(self):
        ret = pd.DataFrame(columns=['x1', 'x2'])
        ret.loc[len(ret)] = 10,20
        ret.loc[len(ret)] = 11,21
        ret.loc[len(ret)] = 12,22
        ret.loc[len(ret)] = 13,23
        ret.loc[len(ret)] = 14,24
        ret.loc[len(ret)] = 15,25
        return ret
    
    @pytest.fixture
    def y_true(self):
        return pd.Series([
            30,
            31,
            32,
            33,
            34,
            35,
        ])

    @pytest.fixture
    def cv_model(self, cv_splits):
        cv_mod = [self.pipeline(0) for i in range(len(cv_splits))]
        return dict(estimator=cv_mod)
    
    def metric_1(self, targ, out):
        return targ.sum() - out.sum()

    def metric_2(self, targ, out):
        return targ.sum() + out.sum()

    def test_single_metric(self, cv_model, cv_splits, X, y_true):
        metric_map = {'Metric_1' : self.metric_1}
        df = get_staged_metrics(cv_model, cv_splits, X, y_true, metric_map)

        assert df.iloc[0].Metric == 'Metric_1'
        assert df.iloc[0].Value == 63
        assert df.iloc[0].Stage == 0
        assert df.iloc[0].Fold == 0
    
        assert df.iloc[1].Metric == 'Metric_1'
        assert df.iloc[1].Value == 29
        assert df.iloc[1].Stage == 1
        assert df.iloc[1].Fold == 0

        assert df.iloc[2].Metric == 'Metric_1'
        assert df.iloc[2].Value == -9
        assert df.iloc[2].Stage == 2
        assert df.iloc[2].Fold == 0
    
        assert df.iloc[3].Metric == 'Metric_1'
        assert df.iloc[3].Value == 66
        assert df.iloc[3].Stage == 0
        assert df.iloc[3].Fold == 1
    
        assert df.iloc[4].Metric == 'Metric_1'
        assert df.iloc[4].Value == 26
        assert df.iloc[4].Stage == 1
        assert df.iloc[4].Fold == 1
    
        assert df.iloc[5].Metric == 'Metric_1'
        assert df.iloc[5].Value == -18
        assert df.iloc[5].Stage == 2
        assert df.iloc[5].Fold == 1


    def test_multiple_metric(self, cv_model, cv_splits, X, y_true):
        metric_map = {'Metric_1' : self.metric_1, 'Metric_2' : self.metric_2}
        df = get_staged_metrics(cv_model, cv_splits, X, y_true, metric_map)

        assert df.iloc[0].Metric == 'Metric_1'
        assert df.iloc[0].Value == 63
        assert df.iloc[0].Stage == 0
        assert df.iloc[0].Fold == 0

        assert df.iloc[1].Metric == 'Metric_2'
        assert df.iloc[1].Value == 123
        assert df.iloc[1].Stage == 0
        assert df.iloc[1].Fold == 0

        assert df.iloc[2].Metric == 'Metric_1'
        assert df.iloc[2].Value == 29
        assert df.iloc[2].Stage == 1
        assert df.iloc[2].Fold == 0

        assert df.iloc[3].Metric == 'Metric_2'
        assert df.iloc[3].Value == 157
        assert df.iloc[3].Stage == 1
        assert df.iloc[3].Fold == 0

        assert df.iloc[4].Metric == 'Metric_1'
        assert df.iloc[4].Value == -9
        assert df.iloc[4].Stage == 2
        assert df.iloc[4].Fold == 0

        assert df.iloc[5].Metric == 'Metric_2'
        assert df.iloc[5].Value == 195
        assert df.iloc[5].Stage == 2
        assert df.iloc[5].Fold == 0

        assert df.iloc[6].Metric == 'Metric_1'
        assert df.iloc[6].Value == 66
        assert df.iloc[6].Stage == 0
        assert df.iloc[6].Fold == 1

        assert df.iloc[7].Metric == 'Metric_2'
        assert df.iloc[7].Value == 138
        assert df.iloc[7].Stage == 0
        assert df.iloc[7].Fold == 1

        assert df.iloc[8].Metric == 'Metric_1'
        assert df.iloc[8].Value == 26
        assert df.iloc[8].Stage == 1
        assert df.iloc[8].Fold == 1

        assert df.iloc[9].Metric == 'Metric_2'
        assert df.iloc[9].Value == 178
        assert df.iloc[9].Stage == 1
        assert df.iloc[9].Fold == 1

        assert df.iloc[10].Metric == 'Metric_1'
        assert df.iloc[10].Value == -18
        assert df.iloc[10].Stage == 2
        assert df.iloc[10].Fold == 1
    
        assert df.iloc[11].Metric == 'Metric_2'
        assert df.iloc[11].Value == 222
        assert df.iloc[11].Stage == 2
        assert df.iloc[11].Fold == 1


    def test_single_pre_processing(self, cv_splits, X, y_true):
        metric_map = {'Metric_1' : self.metric_1}
        cv_model = dict(estimator=[self.pipeline(10) for i in range(len(cv_splits))])

        df = get_staged_metrics(cv_model, cv_splits, X, y_true, metric_map)
        assert df.iloc[0].Value == 83
        assert df.iloc[1].Value == 69
        assert df.iloc[2].Value == 51
        assert df.iloc[3].Value == 86
        assert df.iloc[4].Value == 66
        assert df.iloc[5].Value == 42


    def test_multiple_pre_processing(self, cv_splits, X, y_true):
        metric_map = {'Metric_1' : self.metric_1}

        pipe = ObjectView()
        pipe.steps = [
            ('pre_proc', TestGetStagedMetrics.FakePreProcessing(3)),
            ('pre_proc_2', TestGetStagedMetrics.FakePreProcessing(7)),
            ('estimator', TestGetStagedMetrics.FakePredictor())
        ]
        cv_model = dict(estimator=[copy.deepcopy(pipe) for i in range(len(cv_splits))])

        df = get_staged_metrics(cv_model, cv_splits, X, y_true, metric_map)
        assert df.iloc[0].Value == 83
        assert df.iloc[1].Value == 69
        assert df.iloc[2].Value == 51
        assert df.iloc[3].Value == 86
        assert df.iloc[4].Value == 66
        assert df.iloc[5].Value == 42



"""Test file foor pre_process.py module"""

import pytest
import numpy as np
import pandas as pd
import sklearn.pipeline
from lpsds.utils import keep, ObjectView, to_list, smart_tuple, confusion_matrix_annotation, pretty_title, error_bar_roc, pipeline_split

        
class TestKeep():
    @pytest.fixture
    def df(self):
        data = {}
        data['a'] = [11,22,33,44,55]
        data['b'] = [111,222,333,444,555]
        data['c'] = [True,True,False,True,False]
        data['d'] = ['v1','v2','v3','v4','v5']
        return pd.DataFrame(data)

    def test_remove_index(self, df):
        to_keep = df.a.isin([22,44])
        keep(df, index=to_keep)
        assert len(df) == 2
        assert df.index[0] == 1
        assert df.index[1] == 3

    def test_remove_column(self, df):
        to_keep = df.columns.isin(['a','d'])
        keep(df, columns=to_keep)
        assert len(df.columns) == 2
        assert df.columns[0] == 'a'
        assert df.columns[1] == 'd'
    
    def test_remove_index_and_columns(self, df):
        idx_to_keep = df.a.isin([11,22,55])
        cols_to_keep = df.columns == 'c'
        keep(df, index=idx_to_keep, columns=cols_to_keep)
        
        #Checking index
        assert len(df) == 3
        assert df.index[0] == 0
        assert df.index[1] == 1
        assert df.index[2] == 4
        
        #Checking columns
        assert len(df.columns) == 1
        assert df.columns[0] == 'c'

    def test_inplace_true(self, df):
        idx_to_keep = df.a.isin([33,44])
        ret = keep(df, index=idx_to_keep, inplace=True)
        assert ret is None

    def test_inplace_false(self, df):
        idx_to_keep = df.a.isin([33,44])
        ret = keep(df, index=idx_to_keep, inplace=False)
        
        #Ckecking that ret is not None and that df is unchanged.
        assert ret is not None
        assert len(df) == 5
        assert len(ret) == 2


class TestObjectView:
    """Test ObjectView class"""

    @pytest.fixture
    def obj(self):
        """Default ObjectView object"""
        return ObjectView()

    def test_set_value(self, obj):
        """Test if the obj.var =  value works"""
        obj.my_field = 10
        assert obj['my_field'] == 10

    def test_get_value(self, obj):
        """Test if we can read the value using the notation 'obj.var'"""
        obj['my_field'] = 10
        assert obj.my_field == 10

    def test_attribute_not_found(self, obj):
        """Test if an AttributeError is raised if a required field does not exists."""
        obj.my_field = 10
        with pytest.raises(AttributeError):
            obj.my_other_field
    
    def test_attribute_not_found_for_deletion(self, obj):
        """Test if an AttributeError is raised if a required field does not exists upon deletion."""
        obj.my_field = 10
        with pytest.raises(AttributeError):
            del obj.my_other_field



class TestToList:
    """Test to_list function"""

    def test_value_to_list(self):
        """Test if a value gets converted to a list"""
        ret = to_list(10)
        assert type(ret) is list
        assert len(ret) == 1
        assert ret[0] == 10

    def assert_list_remains_as_list(self):
        """A passed list must be returned unchanged"""
        val = [1]
        ret = to_list(val)
        assert type(ret) is list
        assert len(ret) == 1
        assert ret[0] == 10

    def assert_list_remains_as_tuple(self):
        """A passed tuple must be returned unchanged"""
        val = (10, 20)
        ret = to_list(val)
        assert type(ret) is tuple
        assert len(ret) == 2
        assert ret[0] == 10
        assert ret[1] == 20

    def assert_string_remains_as_string(self):
        """A passed string must be returned unchanged"""
        val = "my test string"
        ret = to_list(val)
        assert type(ret) is str
        assert ret == val


class TestSmartTuple:
    """Test smart_tuple function"""

    def test_single_value_tuple(self):
        val = (10,)
        assert smart_tuple(val) == 10

    def test_multi_value_tuple(self):
        val = (10, 20, 30)
        ret = smart_tuple(val)
        assert type(ret) is tuple
        assert len(ret) == 3
        assert ret[0] == 10
        assert ret[1] == 20
        assert ret[2] == 30



class TestCMAnnotate:

    @pytest.fixture
    def cm(self):
        return np.array([
            [[1,7,20],[40,5,16]],
            [[19,1,9], [10,21,2]],
            [[17,14,15],[26,17,18]],
            [[19,40,21], [32,23,24]]
        ])
    
    @pytest.fixture
    def fmt_str(self):
        return '${mean:.2f}^{{+{e_max:.2f}}}_{{-{e_min:.2f}}}$'
    
    @pytest.fixture
    def seed(self):
        return 23
    
    
    def test_ret_shape(self, cm):
        ret = confusion_matrix_annotation(cm)
        assert len(ret.shape) == 2
        assert ret.shape[0] == 2
        assert ret.shape[1] == 3
    

    def test_difference_false(self, cm, fmt_str, seed):
        ret = confusion_matrix_annotation(cm, use_difference=False, seed=seed)
        assert ret.iloc[0][0] == fmt_str.format(mean=14.00, e_min=5.49, e_max=19.00)
        assert ret.iloc[0][1] == fmt_str.format(mean=15.50, e_min=4.00, e_max=31.75)
        assert ret.iloc[0][2] == fmt_str.format(mean=16.25, e_min=11.75, e_max=20.50)

        assert ret.iloc[1][0] == fmt_str.format(mean=27.00, e_min=15.50, e_max=36.50)
        assert ret.iloc[1][1] == fmt_str.format(mean=16.50, e_min=8.98, e_max=22.00)
        assert ret.iloc[1][2] == fmt_str.format(mean=15.00, e_min=6.00, e_max=22.00)


    def test_difference_true(self, cm, fmt_str, seed):
        ret = confusion_matrix_annotation(cm, use_difference=True, seed=seed)
        assert ret.iloc[0][0] == fmt_str.format(mean=14.00, e_min=8.51, e_max=5.00)
        assert ret.iloc[0][1] == fmt_str.format(mean=15.50, e_min=11.50, e_max=16.25)
        assert ret.iloc[0][2] == fmt_str.format(mean=16.25, e_min=4.5, e_max=4.25)

        assert ret.iloc[1][0] == fmt_str.format(mean=27.00, e_min=11.50, e_max=9.50)
        assert ret.iloc[1][1] == fmt_str.format(mean=16.50, e_min=7.52, e_max=5.50)
        assert ret.iloc[1][2] == fmt_str.format(mean=15.00, e_min=9.00, e_max=7.00)


    def test_fmt_str(self, cm, seed):
        fmt_str = '{mean:.0f}, {e_min:.0f}, {e_max:.0f}'
        ret = confusion_matrix_annotation(cm, use_difference=True, seed=seed, fmt_str=fmt_str)
        assert ret.iloc[0][0] == fmt_str.format(mean=14, e_min=9, e_max=5)
        assert ret.iloc[0][1] == fmt_str.format(mean=16, e_min=12, e_max=16)
        assert ret.iloc[0][2] == fmt_str.format(mean=16, e_min=4, e_max=4)

        assert ret.iloc[1][0] == fmt_str.format(mean=27, e_min=12, e_max=10)
        assert ret.iloc[1][1] == fmt_str.format(mean=16, e_min=8, e_max=6)
        assert ret.iloc[1][2] == fmt_str.format(mean=15, e_min=9, e_max=7)


class TestPrettyTitle:

    @pytest.mark.parametrize(("val", "ret"), [ 
                                                ('auc', 'AUC'),
                                                ('Auc', 'AUC'),
                                                ('aUc', 'AUC'),
                                                ('auC', 'AUC'),
                                                ('AUc', 'AUC'),
                                                ('aUC', 'AUC'),
                                                ('AUC', 'AUC'),
                                                ('specificity', 'Specificity'),
                                                ('Specificity', 'Specificity'),
                                                ('SPECIFICITY', 'Specificity'),
                                            ])
    def test_normal_case(self, val, ret):
        assert pretty_title(val) == ret


    @pytest.mark.parametrize(("val", "ret"), [ 
                                                ('auc', 'Auc'),
                                                ('Auc', 'Auc'),
                                                ('aUc', 'Auc'),
                                                ('auC', 'Auc'),
                                                ('AUc', 'Auc'),
                                                ('aUC', 'Auc'),
                                                ('AUC', 'Auc'),
                                                ('specificity', 'Specificity'),
                                                ('Specificity', 'Specificity'),
                                                ('SPECIFICITY', 'Specificity'),
                                            ])
    def test_new_threshold(self, val, ret):
        assert pretty_title(val, abbrev_threshold=2) == ret


    @pytest.mark.parametrize(("val", "ret"), [ 
                                                ('auc', 'AUC'),
                                                ('Auc', 'AUC'),
                                                ('aUc', 'AUC'),
                                                ('auC', 'AUC'),
                                                ('AUc', 'AUC'),
                                                ('aUC', 'AUC'),
                                                ('AUC', 'AUC'),
                                                ('specificity', 'SPECIFICITY'),
                                                ('Specificity', 'SPECIFICITY'),
                                                ('SPECIFICITY', 'SPECIFICITY'),
                                            ])
    def test_new_threshold_large(self, val, ret):
        assert pretty_title(val, abbrev_threshold=20) == ret




class TestErrorBarRoc:

    @pytest.fixture
    def roc_df(self):
        df = pd.DataFrame(columns=['fold', 'false_positive', 'true_positive', 'threshold'])
        df.loc[len(df)] = 0, 0, 0, 0 # tp = 2*fp
        df.loc[len(df)] = 0, 0.2, 0.4, 0
        df.loc[len(df)] = 0, 0.3, 0.6, 0
        df.loc[len(df)] = 0, 1.0, 2.0, 0
        df.loc[len(df)] = 1, 0, 0, 0 # tp = 10*fp
        df.loc[len(df)] = 1, 0.6, 6, 0
        df.loc[len(df)] = 1, 1, 10, 0
        return df

    @pytest.fixture
    def ret(self, roc_df):
        return error_bar_roc(roc_df, num_points=5)

    def test_ret_size(self, ret):
        assert len(ret) == 10


    def test_col_names(self, ret):
        assert len(ret.columns) == 3
        assert ret.columns[0] == 'false_positive'
        assert ret.columns[1] == 'true_positive'
        assert ret.columns[2] == 'fold'
    
    def test_fold_0(self, ret):
        fold = ret.loc[ret.fold==0]
        assert len(fold) == 5
        assert fold.iloc[0].true_positive == 0
        assert fold.iloc[1].true_positive == 0.5
        assert fold.iloc[2].true_positive == 1
        assert fold.iloc[3].true_positive == 1.5
        assert fold.iloc[4].true_positive == 2


    def test_fold_1(self, ret):
        fold = ret.loc[ret.fold==1]
        assert len(fold) == 5
        assert fold.iloc[0].true_positive == 0
        assert fold.iloc[1].true_positive == 2.5
        assert fold.iloc[2].true_positive == 5.0
        assert fold.iloc[3].true_positive == 7.5
        assert fold.iloc[4].true_positive == 10

    def test_fp_(self, ret):
        fold0 = ret.loc[ret.fold==0]
        assert fold0.iloc[0].false_positive == 0
        assert fold0.iloc[1].false_positive == 0.25
        assert fold0.iloc[2].false_positive == 0.5
        assert fold0.iloc[3].false_positive == 0.75
        assert fold0.iloc[4].false_positive == 1.0


    def test_fp_equal(self, ret):
        fold0 = ret.loc[ret.fold==0]
        fold1 = ret.loc[ret.fold==1]
        assert fold0.iloc[0].false_positive == fold1.iloc[0].false_positive
        assert fold0.iloc[1].false_positive == fold1.iloc[1].false_positive
        assert fold0.iloc[2].false_positive == fold1.iloc[2].false_positive
        assert fold0.iloc[3].false_positive == fold1.iloc[3].false_positive
        assert fold0.iloc[4].false_positive == fold1.iloc[4].false_positive
    


class TestPipelineSplit:

    class PreProc1:
        def transform(self, X):
            return X + 1
    class PreProc2:
        def transform(self, X):
            return X + 10
    class Estimator:
        def predict(self, X):
            return np.array([11,22])


    @pytest.fixture
    def pipe(self):
        return sklearn.pipeline.Pipeline(steps = [
            ('pp1', TestPipelineSplit.PreProc1()),
            ('pp2', TestPipelineSplit.PreProc2()),
            ('est', TestPipelineSplit.Estimator()),
        ])
    
    def test_pre_proc_split(self, pipe):
        pp, mod, x = pipeline_split(pipe)
        assert len(pp) == 2
        assert pp[0][0] == 'pp1'
        assert isinstance(pp[0][1], TestPipelineSplit.PreProc1)
        assert pp[1][0] == 'pp2'
        assert isinstance(pp[1][1], TestPipelineSplit.PreProc2)


    def test_estimator(self, pipe):
        pp, mod, x = pipeline_split(pipe)
        assert mod[0] == 'est'
        assert isinstance(mod[1], TestPipelineSplit.Estimator)


    def test_x_none(self, pipe):
        pp, mod, x = pipeline_split(pipe)
        assert x is None


    def test_x(self, pipe):
        X = np.array([[1,2],
                      [3,4]])
        pp, mod, x = pipeline_split(pipe, X)
        assert X.shape[0] == 2
        assert X.shape[1] == 2
        assert x[0][0] == 12
        assert x[0][1] == 13
        assert x[1][0] == 14
        assert x[1][1] == 15



import pytest
import pandas as pd
import numpy as np
from lpsds.preprocessing import StandardNaN, BooleanEncode, SmartFloatCasting, drop_null_cols


class TestDropNullCols:
    """Test drop_null_cols"""

    @pytest.fixture
    def df(self):
        """Test dataset"""

        return pd.DataFrame({
            'a' : [1,2,3,None],
            'b' : [111,np.nan,np.nan, None],
            'c' : [11,22,np.nan, None],
            'd' : [None,np.nan,np.nan, None],
        })
    
    def test_operation(self, df):
        """Test normal behavior"""
        ret = drop_null_cols(df, threshold=0.5)
        cols = ret.columns
        assert 'a' in cols
        assert 'b' not in cols
        assert 'c' in cols
        assert 'd' not in cols


    def test_inplace_true(self, df):
        """Test inplace True"""
        drop_null_cols(df, threshold=0.5, inplace=True)
        cols = df.columns
        assert 'a' in cols
        assert 'b' not in cols
        assert 'c' in cols
        assert 'd' not in cols

    
    def test_inplace_false(self, df):
        """Test inplace False"""
        drop_null_cols(df, threshold=0.5, inplace=False)
        cols = df.columns
        assert 'a' in cols
        assert 'b' in cols
        assert 'c' in cols
        assert 'd' in cols
    
    def test_std_nan_true(self):
        """Test NaN standartization"""
        df = pd.DataFrame({
            'a' : ['1','','NA','None'],
            'b' : [111,np.nan,np.nan, None],
            'c' : [11,22,np.nan, None],
            'd' : [None,'nan','val', None],
        })
        ret = drop_null_cols(df, threshold=0.5, std_nan=True)
        cols = ret.columns
        assert 'a' not in cols
        assert 'b' not in cols
        assert 'c' in cols
        assert 'd' not in cols

    def test_std_nan_false(self):
        """Test NaN standartization"""
        df = pd.DataFrame({
            'a' : ['1','','NA','None'],
            'b' : [111,np.nan,np.nan, None],
            'c' : [11,22,np.nan, None],
            'd' : [None,'nan','val', None],
        })
        ret = drop_null_cols(df, threshold=0.5, std_nan=False)
        cols = ret.columns
        assert 'a' in cols
        assert 'b' not in cols
        assert 'c' in cols
        assert 'd' in cols



class TestBooleanEncode:
    """Test BooleanEncode class"""

    @pytest.fixture
    def bool_map(self):
        """boolean map"""
        return {
            'pos' : +1,
            'neg' : -1,
            'None' : 0,
        }
    
    @pytest.fixture
    def df(self):
        """Test dataset"""
        return pd.DataFrame({
            'a' : ['pos', 'neg', 'None'],
            'b' : ['pos', 'pos', 'neg']
        })


    def test_more_3_values(self):
        """
        ValueError must be raised 
        if the map has mopre than 3 distinct values for items
        """
        bool_map = dict(
            pos=+1,
            neg=-1,
            other=0,
            another_pos=+2,
        )
        with pytest.raises(ValueError):
            BooleanEncode(bool_map)

    def test_more_3_keys(self):
        """
        No error must occur if more than 3 keys are given as long as
        values are no more than 3
        """
        bool_map = dict(
            pos=+1,
            neg=-1,
            other=0,
            another_pos=+1,
        )
        BooleanEncode(bool_map)
        assert True

    def test_normal_operation(self, bool_map, df):
        """Test normal operation"""
        obj = BooleanEncode(bool_map)
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == +1
        assert ret.a.iloc[1] == -1
        assert ret.a.iloc[2] == 0

        assert ret.b.iloc[0] == +1
        assert ret.b.iloc[1] == +1
        assert ret.b.iloc[2] == -1


    def test_normal_operation_boolean(self):
        """Test normal operation"""
        test_map = {
            True : +1,
            False : -1,
            np.nan : 0,
        }

        df = pd.DataFrame({
            'a' : [True, False, np.nan],
            'b' : [True, True, False]
        })

        obj = BooleanEncode(test_map)
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == +1
        assert ret.a.iloc[1] == -1
        assert ret.a.iloc[2] == 0

        assert ret.b.iloc[0] == +1
        assert ret.b.iloc[1] == +1
        assert ret.b.iloc[2] == -1


    def test_inplace_true(self, bool_map, df):
        """Test inplace True"""
        obj = BooleanEncode(bool_map, inplace=True)
        obj.fit_transform(df)
        assert df.a.iloc[0] == +1
        assert df.a.iloc[1] == -1
        assert df.a.iloc[2] == 0

        assert df.b.iloc[0] == +1
        assert df.b.iloc[1] == +1
        assert df.b.iloc[2] == -1

    def test_inplace_false(self, bool_map, df):
        """Test inplace False"""
        obj = BooleanEncode(bool_map, inplace=False)
        obj.fit_transform(df)
        assert df.a.iloc[0] == 'pos'
        assert df.a.iloc[1] == 'neg'
        assert df.a.iloc[2] == 'None'

        assert df.b.iloc[0] == 'pos'
        assert df.b.iloc[1] == 'pos'
        assert df.b.iloc[2] == 'neg'

    def test_dtype(self, bool_map, df):
        """Test dtype override"""
        obj = BooleanEncode(bool_map, dtype=np.int8)
        ret = obj.fit_transform(df)
        assert str(ret.a.dtype) == 'int8'
        assert str(ret.b.dtype) == 'int8'





class TestStandardNaN:
    """Testing StandardNaN class"""

    @pytest.fixture
    def obj(self):
        """Test object"""
        return StandardNaN()

    def test_simple_case(self, obj):
        """Test simple behavior"""

        df = pd.DataFrame({
            'a' : ['aaa', 'None', 'NA'], 
            'b' : ['', 'bbb', 'NAN'],
        })
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == 'aaa'
        assert np.isnan(ret.a.iloc[1])
        assert np.isnan(ret.a.iloc[2])

        assert np.isnan(ret.b.iloc[0])
        assert ret.b.iloc[1] == 'bbb'
        assert np.isnan(ret.b.iloc[2])


    def test_nan_rep_addition(self):
        """Test adding new nan rep"""

        df = pd.DataFrame({
            'a' : ['aaa', 'None', 'ccc'], 
            'b' : ['', 'bbb', 'NAN'],
        })

        obj = StandardNaN(additional_nan_rep=['bbb', 'ccc'])
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == 'aaa'
        assert np.isnan(ret.a.iloc[1])
        assert np.isnan(ret.a.iloc[2])

        assert np.isnan(ret.b.iloc[0])
        assert np.isnan(ret.b.iloc[1])
        assert np.isnan(ret.b.iloc[2])


    def test_std_nan_override(self):
        """Testing setting a new NaN override"""

        df = pd.DataFrame({
            'a' : ['aaa', 'None', 'ccc'], 
            'b' : ['', 'bbb', 'NAN'],
        })

        std_val = 'NEW'
        obj = StandardNaN(additional_nan_rep=['aaa'], std_nan_val=std_val)
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == std_val
        assert ret.a.iloc[1] == std_val
        assert ret.a.iloc[2] == 'ccc'

        assert ret.b.iloc[0] == std_val
        assert ret.b.iloc[1] == 'bbb'
        assert ret.b.iloc[2] == std_val


    def test_case_insensitiveness(self, obj):
        """Test whether the transformation is case insensitive."""

        df = pd.DataFrame({
            'a' : ['na', 'nA', 'NA'], 
            'b' : ['nan', 'Nan', 'NAN'],
            'c' : ['', ' ', '   '],
            'd' : ['<NA>', '<na>', '<Na>'],
            'e' : ['não informado', 'Não Informado', 'Não InFoRmAdO'],
            'f' : ['None', 'nONe', 'NONE'],
        })
        res = obj.fit_transform(df)

        for idx, c in df.iterrows():
            for col in df.columns:
                assert np.isnan(np.nan)


    def test_numeric_columns(self, obj):
        """Test if the class can handle dataframe with numeric columns"""

        df = pd.DataFrame({
            'a' : [1, 2, np.nan], 
            'b' : ['nan', 'To Keep', '<na>'],
            'c' : [4.1, np.nan, 4.3],
            'd' : ['To Keep', '<na>', '<Na>'],
        })
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == 1
        assert ret.a.iloc[1] == 2
        assert np.isnan(ret.a.iloc[2])

        assert np.isnan(ret.b.iloc[0])
        assert ret.b.iloc[1] == 'To Keep'
        assert np.isnan(ret.b.iloc[2])

        assert ret.c.iloc[0] == 4.1
        assert np.isnan(ret.c.iloc[1])
        assert ret.c.iloc[2] == 4.3

        assert ret.d.iloc[0] == 'To Keep'
        assert np.isnan(ret.d.iloc[1])
        assert np.isnan(ret.d.iloc[2])

    def test_inplace_true(self):
        """Test inplace = True"""

        df = pd.DataFrame({
            'a' : ['To Keep', '<na>'], 
            'b' : ['None', 'To Keep'], 
        })

        obj = StandardNaN(inplace=True)
        obj.fit_transform(df)

        assert df.a.iloc[0] == 'To Keep'
        assert np.isnan(df.a.iloc[1])
        assert np.isnan(df.b.iloc[0])
        assert df.b.iloc[1] == 'To Keep'


    def test_inplace_false(self):
        """Test inplace = False"""

        df = pd.DataFrame({
            'a' : ['To Keep', '<na>'], 
            'b' : ['None', 'To Keep'], 
        })

        obj = StandardNaN(inplace=False)
        obj.fit_transform(df)

        assert df.a.iloc[0] == 'To Keep'
        assert df.a.iloc[1] == '<na>'
        assert df.b.iloc[0] == 'None'
        assert df.b.iloc[1] == 'To Keep'



class TestSmartFloatCasting:
    """Test SmartFloatCasting"""
    
    @pytest.fixture
    def df(self):
        """Test dataset"""
        return pd.DataFrame({
            'a' : ['1,23', '-5',],
            'b' : ['-23.53', '0',],
        })

    def test_normal_operation(self, df):
        """test normal operation"""
        obj = SmartFloatCasting()
        ret = obj.fit_transform(df)
        assert ret.a.iloc[0] == 1.23
        assert ret.a.iloc[1] == -5
        assert ret.b.iloc[0] == -23.53
        assert ret.b.iloc[1] == 0

    def test_inplace_true(self, df):
        """test inplace True"""
        obj = SmartFloatCasting(inplace=True)
        obj.fit_transform(df)
        assert df.a.iloc[0] == 1.23
        assert df.a.iloc[1] == -5
        assert df.b.iloc[0] == -23.53
        assert df.b.iloc[1] == 0

    def test_inplace_false(self, df):
        """test inplace False"""
        obj = SmartFloatCasting(inplace=False)
        obj.fit_transform(df)
        assert df.a.iloc[0] == '1,23'
        assert df.a.iloc[1] == '-5'
        assert df.b.iloc[0] == '-23.53'
        assert df.b.iloc[1] == '0'

    def test_dtype(self, df):
        """test different dtype"""
        obj = SmartFloatCasting(dtype=np.float32)
        ret = obj.fit_transform(df)
        assert str(ret.a.dtype) == 'float32'
        assert str(ret.b.dtype) == 'float32'



import numpy as np
import pandas as pd
from typing import Union


def get_fold_data(cv_model: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], fold_idx: int) -> tuple:
     """"
     def get_fold_data(cv_model: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], fold_idx: int) -> tuple

     Returns the data corresponding to a given cross validation fold.

     Input parameters:
       - cv_model: a dictionary with structure equivalent to the one returned by sklearn.model_selection.cross_validate.
       - cv_splits: a list where each item is a tuple containing the indexes to be used for training and testing in
                    see sklearn.model_selection.KFold, for isntance, for details.
       - X: a DataFrame containing **ALL** input samples used for model development and testing. It should be the same X as
            the one passed to sklearn.model_selection.cross_validate.
       - y_true: a Series containing **ALL** target samples used for model development and testing. It should be the same y_true as
            the one passed to sklearn.model_selection.cross_validate.
       - fold_idx: the index of the cv fold you want to collect the data.
    
     Returns:
       - X_train: the input set used for training in the desired fold.
       - X_test: the input set used for testing in the desired fold.
       - y_train: the target set used for training in the desired fold.
       - y_test: the target set used for testing in the desired fold.
       - model: the operation model yielded by the desired fold.
     """
     trn_idx, tst_idx = cv_splits[fold_idx]

     if hasattr(X, 'iloc'):
          X_train = X.iloc[trn_idx]
          X_test = X.iloc[tst_idx]
     else:
          X_train = X[trn_idx,]
          X_test = X[tst_idx,]

     if hasattr(y_true, 'iloc'):
          y_train = y_true.iloc[trn_idx]
          y_test = y_true.iloc[tst_idx]
     else:
          y_train = y_true[trn_idx,]
          y_test = y_true[tst_idx,]
    
     model = cv_model['estimator'][fold_idx]
    
     return X_train, X_test, y_train, y_test, model



def get_operation_model(cv_model: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], metric:str='test_sp', less_is_better:bool=False) -> tuple:
    """"
    get_operation_model(cv_model: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], metric:str='test_sp', less_is_better:bool=False) -> tuple

    Returns the operation model and its corresponding data. For that, it finds
    the model yielding the best metric over its corresponding testing set.

    Input parameters:
      - cv_model: a dictionary with structure equivalent to the one returned by sklearn.model_selection.cross_validate.
      - cv_splits: a list where each item is a tuple containing the indexes to be used for training and testing in
                   see sklearn.model_selection.KFold, for isntance, for details.
      - X: a DataFrame containing **ALL** input samples used for model development and testing. It should be the same X as
           the one passed to sklearn.model_selection.cross_validate.
      - y_true: a Series containing **ALL** target samples used for model development and testing. It should be the same y_true as
           the one passed to sklearn.model_selection.cross_validate.
      - metric: a key within cv_model containing the metric to use to compare cv folds.
      - less_is_better: if True, the best fold will be the one where the passed metric is the lowest (MSE, for instance). The highest
                        will be returned if this value is set to True (Accuracy, for instance).
    
    Returns:
      - best_fold: the id of the fold yielding the operation model.
      - X: the input set used for testing in best_fold.
      - y_true: the target set used for testing in best_fold.
      - model: the operation model yielded by best_fold
    """
    best_fold = cv_model[metric].argmin() if less_is_better else cv_model[metric].argmax()
    _, X_tst, _, y_tst, model = get_fold_data(cv_model, cv_splits, X, y_true, best_fold)
    return best_fold, X_tst, y_tst, model


def get_staged_metrics(cv: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], metrics_map: dict, **kwargs) -> pd.DataFrame:
     """
     def get_staged_metrics(cv: dict, cv_splits: list, X: Union[pd.DataFrame, np.array], y_true: Union[pd.Series, np.array], metrics_map: dict, **kwargs) -> pd.DataFrame:

     Evaluates a set of metrics for each stage in an enseble model taking into consideration the cross-validation folds employed for the model development.

     Input parameters:
      - cv: a dictionary with structure equivalent to the one returned by sklearn.model_selection.cross_validate.
      - cv_splits: a list where each item is a tuple containing the indexes to be used for training and testing in
                   see sklearn.model_selection.KFold, for isntance, for details.
      - X: a DataFrame containing **ALL** input samples used for model development and testing. It should be the same X as
           the one passed to sklearn.model_selection.cross_validate.
      - y_true: a Series containing **ALL** target samples used for model development and testing. It should be the same y_true as
                the one passed to sklearn.model_selection.cross_validate.
      - metrics_map: a map where key is a string label identifying the metric to be evaluated at each stage, and value is a reference
                     to a function in the format metric_function(y_pred, y_true, **kwargs) -> float that will be responsible to evaluate
                     the desired metric for each stage of the ensemble model for each cv-fold.

     Returns: a pandas.DataFrame with the following columns:
       - Metric: the metric name (given by the keys in metrics_map)
       - Fold: the cv-fold ID where the metric is being evaluated.
       - Stage: the number of stages considered when evaluating the metric.
       - Value: the achieved metric value.
     """

     aux_df_list = []
     num_folds = len(cv_splits)
     for fold in range(num_folds):
         #Getting the test data and model for the fold
         _, in_tst, _, targ_tst, model = get_fold_data(cv, cv_splits, X, y_true, fold)

         #Pipelines do not implemet staged_predict. As a result, we need
         #to apply the pre-processing manually. Probably there is a better way to do this...
         for _, pp in model.steps[:-1]:
             if hasattr(pp, 'transform'):
                 in_tst = pp.transform(in_tst)

         #Collecting just the final estimator (i.e. the model being applied right after all the pre-processing pipeline).
         estimator = model.steps[-1][1]

         #Calculating the accuracy for each class w.r.t the number of stages.
         for stage, out_tst in enumerate(estimator.staged_predict(in_tst)):
             for metric_name, metric_function in metrics_map.items():
                 metric_value = metric_function(targ_tst, out_tst, **kwargs)
                 #Saving the results in an auxiliary dataframe
                 aux_df = pd.DataFrame({
                     'Metric' : [metric_name],
                     'Value' : [metric_value],
                     'Stage' : [stage],
                     'Fold' : [fold],
                 })
                 aux_df_list.append(aux_df)

     #Creating the final dataframe in a format suitable for seaborn (long-format)
     return pd.concat(aux_df_list)
 


import io
import os
import pandas as pd
import s3fs
from contextlib import contextmanager


class FileHandler():
    def __init__(self, bucket):
        self.bucket = bucket
    def get_full_path(self, dir_name, prefix='', fname=''):
        raise NotImplemented
    def fopen(self, fname, mode):
        raise NotImplemented
        
        
        
class S3FileHandler(FileHandler):
    def __init__(self, bucket):
        super().__init__(bucket)

    def get_full_path(self, dir_name='', prefix='', fname=''):
        return os.path.join(f's3://{self.bucket}', dir_name, prefix, fname)
    
    @contextmanager
    def fopen(self, fname, mode):
        fs = s3fs.S3FileSystem()
        fobj = fs.open(fname, mode)
        yield fobj
        fobj.close()

        
class LocalFileHandler(FileHandler):    
    def __init__(self, bucket):
        super().__init__(bucket)

    def get_full_path(self, dir_name='', prefix='', fname=''):
        """
        get_full_path(dir_name, prefix='', fname='')

        Creates a file path as dir_"name/prefix/fname".
        Creates "dir_name/prefix" if it does not exists.
        Returns: "dir_name/prefix/fname"
        """
        base_path =  os.path.join(self.bucket, dir_name, prefix)

        #If both dir_name and prefix are empty, we can assume
        #fname contains the full path.
        if (dir_name == '') and (prefix == ''):
            dir_name, fname = os.path.split(fname)

        #Creating full absolute path.
        path = os.path.abspath(os.path.join(self.bucket, dir_name, prefix))
        #If the path does not exist, try using relative path instead
        if not os.path.exists(path): path = os.path.join(self.bucket, dir_name, prefix)
        #If the path still does not exist, we create it
        if not os.path.exists(path): os.makedirs(path)
        #Appending the file name (if existent) and returning.
        return os.path.join(path, fname)
    
    @contextmanager
    def fopen(self, fname, mode):
        f = open(fname, mode)
        yield f
        f.close()



"""
Tools for model handling
"""

from sklearn.metrics import mean_squared_error
import numpy as np
import pandas as pd
from typing import Union


def get_operation_model(model, X, y, metric=mean_squared_error, **kargv):
    """
    def get_operation_model(model, X, y, metric=mean_squared_error, *argv)

    Selects an operation model by running the entire dataset for each fold
    and picking the one with the best metric.

    The function always assume that the best value for a metric is its minimum value
    (MSE, for instance). If you use a metric like R2, where the higher the better, you must
    multiply such metric by -1.

    Input parameters:
        - model: a model structure as returned by sklearn cross_validation.
        - X: the input dataset
        - y: the target values
        - metric: a metric to be used (MSE, R2, etc.)
        - **kargv: passed down to the metric function.
    """
    res = np.array([metric(y, mod.predict(X), **kargv) for mod in model['estimator']])
    op_idx = np.argmin(res)
    return model['estimator'][op_idx]


def create_validation_dataset(model, X, y, proba: bool=True, copy: bool=True):
    """
    def create_validation_dataset(model, X, y, proba: bool=True, copy: bool=True)

    Create a validation dataset suitable for validate your model
    when deployed to another operation environment.

    Input parameters are:
        - model: a sklearn-like model (pipeline, skorch, etc.)
        - X: the input dataset
        - y: the targets of X
        - proba: if True, will call model.predict_proba. Otherwise, model.predict
        - copy: whether to return a copy of the passed dataset (both X and y).
    
    Returns:
        A new dataset containing 2 new columns:
            - y_true: containing the targets in y.
            - y_pred: the model output for each sample in X.
    """

    if copy: X = X.copy()
    X['y_true'] = y
    X['y_pred'] = model.predict_proba(X) if proba else model.predict(X)
    return X


def get_input_variables_description(X):
    """
    Create a descriptive pandas telling the following info from X:
      - The order of each column (Inpout Order)
      - Column names (Input Name)
      - Column types (Input Type)
    """
    input_description = X.dtypes.to_frame().reset_index()
    input_description.index.name = 'Input Order'
    input_description.columns = ['Input Name', 'Input Type']
    return input_description


def feature_importances(model, X :Union[pd.DataFrame, np.ndarray], y_true :Union[pd.Series, np.ndarray], suppressing_function=np.mean,
                        metric_function=mean_squared_error, comparison_function=np.subtract) -> pd.DataFrame:
    """
        def feature_importances(model, X :Union[pd.DataFrame, np.ndarray], y_true :Union[pd.Series, np.ndarray], suppressing_function=np.mean,
                                metric_function=mean_squared_error, comparison_function=np.subtract) -> pd.DataFrame:
    
    Calculate feature importance using variables suppressing method. The function will suppress
    ove variable at a time and calculate the difference in a provided metric w.r.t using all features.
    The higher the difference, the higher the variable importance.

    Input:
      - model: a trained model defining the predict_proba method.
      - X: model input set.
      - y_true: true labes of X.
      - suppressing_function: the function to be used to suppress the feature. It must receive the feature and
                              must return a new unique value to represent the feature. Default to arithmetic mean.
      - metric_function: the metric to be used to evaluate model's performance. Must receive y_true and y_pred and return a scalar.
      - comparison_function: a function that will compare the metric obtained when suppressing a metric and
                             the baseline (metric value when no feature is suppressed). Must receive 2 scalars and return a scalar.
    
    Returns: a pandas DataFrame where the index is the feature name and the column ('improtance') is the metric deviation
              after suppressing the given feature.
    """

    baseline = metric_function(y_true, model.predict(X))

    if isinstance(X, np.ndarray):
        X = pd.DataFrame(X, columns=np.arange(X.shape[1]))

    ret = {}
    for c in X.columns:
        aux = X.copy()
        aux[c] = suppressing_function(aux[c])
        relev_metric = metric_function(y_true, model.predict(aux))
        ret[c] = comparison_function(baseline, relev_metric)
    
    ret_df = pd.DataFrame.from_dict(ret, orient='index', columns=['importance'])
    ret_df.index.set_names('feature', inplace=True)
    return ret_df



"""model module tests"""

import pytest
import pandas as pd
import numpy as np
from lpsds.model import get_operation_model, create_validation_dataset, feature_importances


class DummyModel:
    """Dummy model for tests"""
    def __init__(self,fake_output, id):
        """constructor"""
        self.fake_output = fake_output
        self.id = id

    def predict(self, X):
        """predict emulator"""
        ret = np.ones(self.fake_output.shape)
        #Cutting in zero
        ret[self.fake_output<0] = -1.
        return ret

    def predict_proba(self, X):
        """predict_proba emulator"""
        return self.fake_output


class TestGetOperationModel:
    """Tests get_operation_model"""

    @pytest.fixture
    def X(self):
        """Fake X"""
        return np.random.randn(5,2)
    
    @pytest.fixture
    def y_true(self):
        """Fake y"""
        return np.array([0.2, -1, 0, -0.5, 0.9])
    
    @pytest.fixture
    def models(self):
        """Dummy models"""
        mod1 = DummyModel(np.array([0.2, +1, 0, -0.5, -0.9]), id='model1') #2 wrongs (+1 and -0.9)
        mod2 = DummyModel(np.array([0.2, -1, -0.1, -0.5, 0.9]), id='model2') #1 wrongs 9-0.1
        mod3 = DummyModel(np.array([0.2, +1, 0, +0.5, -0.9]), id='model3') #3 wrongs (+1, +0.5 and -0.9)
        return {'estimator' : [mod1, mod2, mod3]}

    @pytest.fixture(autouse=True)
    def global_var(self):
        pytest.iteration_number = 0

    def dummy_metric(self, y_pred, y_true, **kwargs):
        """
        Dummy metric function.
        """
        mod_idx_to_return = kwargs.get('mod_idx', 0)
        ret = -1 if pytest.iteration_number == mod_idx_to_return else +1
        pytest.iteration_number += 1
        return ret

    
    def test_default_case(self, models, X, y_true):
        """Test behavior with default values"""
        mod = get_operation_model(models, X, y_true)
        assert mod.id == 'model2'

    def test_metric_override(self, models, X, y_true):
        """Test if overriding metrics is working"""
        self.metric_idx = 0
        mod = get_operation_model(models, X, y_true, metric=self.dummy_metric)
        assert mod.id == 'model1'

    def test_metrics_kwargs(self, models, X, y_true):
        """Test kwargs for metric function"""
        self.metric_idx = 0
        mod = get_operation_model(models, X, y_true, metric=self.dummy_metric, mod_idx=2)
        assert mod.id == 'model3'



class TesteCreateValidationDataset:
    """Tests create_validation_dataset"""

    @pytest.fixture
    def X(self):
        """Fake X"""
        return pd.DataFrame({
            'a' : [11, 21],
            'b' : [12, 22],
            })
    
    @pytest.fixture
    def y_true(self):
        """Fake y"""
        return np.array([0.2, -1])

    @pytest.fixture
    def y_pred(self):
        """Fake y"""
        return np.array([0.1, -0.8])

    @pytest.fixture
    def model(self, y_pred):
        """Dummy models"""
        return DummyModel(y_pred, id='model')


    def test_columns_order(self, model, X, y_true):
        """Test whether the returned columns are OK"""
        ret = create_validation_dataset(model, X, y_true, proba=False)
        assert ret.columns[0] == 'a'
        assert ret.columns[1] == 'b'
        assert ret.columns[2] == 'y_true'
        assert ret.columns[3] == 'y_pred'

    def test_values(self, model, X, y_true):
        """Test whether the returned values are OK"""
        ret = create_validation_dataset(model, X, y_true, proba=False)
        assert ret.a.iloc[0] == 11
        assert ret.a.iloc[1] == 21
        assert ret.b.iloc[0] == 12
        assert ret.b.iloc[1] == 22
        assert ret.y_true.iloc[0] == 0.2
        assert ret.y_true.iloc[1] == -1
        assert ret.y_pred.iloc[0] == +1
        assert ret.y_pred.iloc[1] == -1


    def test_predict_proba(self, model, X, y_true):
        """Tests predic probability case"""
        ret = create_validation_dataset(model, X, y_true, proba=True)
        assert ret.a.iloc[0] == 11
        assert ret.a.iloc[1] == 21
        assert ret.b.iloc[0] == 12
        assert ret.b.iloc[1] == 22
        assert ret.y_true.iloc[0] == 0.2
        assert ret.y_true.iloc[1] == -1
        assert ret.y_pred.iloc[0] == 0.1
        assert ret.y_pred.iloc[1] == -0.8

    def test_copy_true(self, model, X, y_true):
        """Test copy = True case for X and y"""
        create_validation_dataset(model, X, y_true, copy=True)
        assert 'y_true' not in X.columns
        assert 'y_pred' not in X.columns
        assert X.a.iloc[0] == 11
        assert X.a.iloc[1] == 21
        assert X.b.iloc[0] == 12
        assert X.b.iloc[1] == 22

    def test_copy_false(self, model, X, y_true):
        """Test copy = False case for X and y"""
        create_validation_dataset(model, X, y_true, copy=False)
        assert X.a.iloc[0] == 11
        assert X.a.iloc[1] == 21
        assert X.b.iloc[0] == 12
        assert X.b.iloc[1] == 22
        assert X.y_true.iloc[0] == 0.2
        assert X.y_true.iloc[1] == -1
        assert X.y_pred.iloc[0] == 0.1
        assert X.y_pred.iloc[1] == -0.8


class TestFeatureImportance:
    
    class Model:
        def predict(self, X):
            ret = X.sum(axis=1)
            return ret if isinstance(ret, np.ndarray) else ret.to_numpy()

    @pytest.fixture
    def model(self):
        return TestFeatureImportance.Model()
    
    @pytest.fixture
    def X(self):
        return pd.DataFrame({'a': [3, 11], 'b' : [20, 30]})
    
    @pytest.fixture
    def y_true(self):
        return pd.Series([22, 45])
    
    def test_columns(self, model, X, y_true):
        relev = feature_importances(model, X, y_true)
        assert relev.index.name == 'feature'
        assert relev.columns[0] == 'importance'

    def test_rows(self, model, X, y_true):
        relev = feature_importances(model, X, y_true)
        assert relev.index[0] == 'a'
        assert relev.index[1] == 'b'

    def test_shape(self, model, X, y_true):
        relev = feature_importances(model, X, y_true)
        assert relev.shape[0] == 2
        assert relev.shape[1] == 1

    def test_positive_case(self, model, X, y_true):
        relev = feature_importances(model, X, y_true)
        assert relev.loc['a', 'importance'] == -36
        assert relev.loc['b', 'importance'] == -50

    @pytest.mark.parametrize(("x", "y"), [ 
                                (pd.DataFrame({'a': [3, 11], 'b' : [20, 30]}),            pd.Series([22, 45])),
                                (pd.DataFrame({'a': [3, 11], 'b' : [20, 30]}),            pd.Series([22, 45]).to_numpy()),
                                (pd.DataFrame({'a': [3, 11], 'b' : [20, 30]}).to_numpy(), pd.Series([22, 45])),
                                (pd.DataFrame({'a': [3, 11], 'b' : [20, 30]}).to_numpy(), pd.Series([22, 45]).to_numpy()),
                            ])
    def test_type_combinations(self, model, x,y):
        relev = feature_importances(model, x, y)
        assert relev.iloc[0].importance == -36
        assert relev.iloc[1].importance == -50